"""
Extension for File to handle xxx files
"""
class FileXXX:
    """
    Extension for File to handle xxx files
    """
    @classmethod
    def _write_xxx(cls, textIO, obj):
        """
        Method that does the writing for this filetype.
        Add documentation to File.write()

        :param generalfile.base.classfile.File cls:
        :param textIO:
        :param obj:
        :return:
        """
        path = cls.toPath(textIO.name)
        textIO.write(str(obj))
        return obj

    @classmethod
    def _read_xxx(cls, textIO):
        """
        Method that does the reading for this filetype.
        Add documentation to File.read()

        :param generalfile.base.classfile.File cls:
        :param textIO:
        :return:
        """
        return textIO.read()

# Replace xxx with filetype, has to match





"""Just contains class for File handling tests"""

from generalfile.base.classpath import Path
from generalfile.base.classfile import File


class SetUpWorkDir:
    """Class to set up working dir for tests, File extensions import this class."""
    workingDir = Path(__file__).getParent(2).addPath("tests")
    if not workingDir.endsWithPath("test/tests"):
        raise EnvironmentError(f"Failed setting correct working dir, should be ..test/tests but it's {workingDir}")

    @classmethod
    def activate(cls):
        """Set working dir and clear it after it's made sure it's correct path."""
        File.setWorkingDir(cls.workingDir)
        if File.getWorkingDir().endsWithPath("tests"):
            File.clearFolder("", delete=True)





"""
Tests for generalfile module

Exceptions raised directly by function should be specific Exceptions
Exceptions raised directly by indirect function should be generic Exception
"""
import unittest
import multiprocessing as mp
import time


from generalfile.base.classpath import Path
from generalfile.base.classfile import File
from test.base.setUpWorkDir import SetUpWorkDir

class FileTest(unittest.TestCase):
    def setUp(self):
        """Set working dir and clear folder"""
        SetUpWorkDir.activate()

    def test_toPath(self):
        File.write("exists.txt")
        File.createFolder("folderExists")

        self.assertRaises(AttributeError,       File.toPath, path=File.getAbsolutePath("folderDoesntExist"),  requireFiletype=True)
        self.assertRaises(AttributeError,       File.toPath, path=File.getAbsolutePath("folderDoesntExist"),  requireFiletype=True,     requireExists=True)
        self.assertRaises(AttributeError,       File.toPath, path=File.getAbsolutePath("doesntExist.txt"),    requireFiletype=False)
        self.assertRaises(AttributeError,       File.toPath, path=File.getAbsolutePath("doesntExist.txt"),    requireFiletype=False,    requireExists=True)

        self.assertRaises(AttributeError,       File.toPath, path="folderExists",       requireFiletype=True)
        self.assertRaises(AttributeError,       File.toPath, path="folderDoesntExist",  requireFiletype=True)
        self.assertRaises(AttributeError,       File.toPath, path="folderExists",       requireFiletype=True,   requireExists=True)
        self.assertRaises(AttributeError,       File.toPath, path="folderDoesntExist",  requireFiletype=True,   requireExists=True)
        self.assertRaises(AttributeError,       File.toPath, path="exists.txt",         requireFiletype=False)
        self.assertRaises(AttributeError,       File.toPath, path="doesntExist.txt",    requireFiletype=False)
        self.assertRaises(AttributeError,       File.toPath, path="exists.txt",         requireFiletype=False,   requireExists=True)
        self.assertRaises(AttributeError,       File.toPath, path="doesntExist.txt",    requireFiletype=False,   requireExists=True)

        self.assertRaises(TypeError,            File.toPath, path="exists.txt",         requireFiletype="tsv")
        self.assertRaises(TypeError,            File.toPath, path="doesntExist.txt",    requireFiletype="tsv")
        self.assertRaises(TypeError,            File.toPath, path="exists.txt",         requireFiletype="tsv",  requireExists=True)
        self.assertRaises(TypeError,            File.toPath, path="doesntExist.txt",    requireFiletype="tsv",  requireExists=True)

        self.assertRaises(FileNotFoundError,    File.toPath, path="folderDoesntExist",                          requireExists=True)
        self.assertRaises(FileNotFoundError,    File.toPath, path="doesntExist.txt",                            requireExists=True)
        self.assertRaises(FileNotFoundError,    File.toPath, path="folderDoesntExist",  requireFiletype=False,  requireExists=True)
        self.assertRaises(FileNotFoundError,    File.toPath, path="doesntExist.txt",    requireFiletype=True,   requireExists=True)
        self.assertRaises(FileNotFoundError,    File.toPath, path="doesntExist.txt",    requireFiletype="txt",  requireExists=True)

        self.assertRaises(FileExistsError,      File.toPath, path="folderExists",                               requireExists=False)
        self.assertRaises(FileExistsError,      File.toPath, path="exists.txt",                                 requireExists=False)
        self.assertRaises(FileExistsError,      File.toPath, path="Exists.txt",                                 requireExists=False)
        self.assertRaises(FileExistsError,      File.toPath, path="folderExists",       requireFiletype=False,  requireExists=False)
        self.assertRaises(FileExistsError,      File.toPath, path="exists.txt",         requireFiletype=True,   requireExists=False)
        self.assertRaises(FileExistsError,      File.toPath, path="exists.txt",         requireFiletype="txt",  requireExists=False)

        path = Path("test")
        self.assertIsNot(File.toPath(path), path)
        self.assertIsNot(File.toPath("test"), File.toPath("test"))
        self.assertEqual(File.toPath("test"), File.toPath("test"))

        self.assertEqual(File.toPath("test.txt", requireFiletype=True), "test.txt")
        self.assertEqual(File.toPath("test.txt", requireFiletype="txt"), "test.txt")
        File.write("test.txt")
        self.assertEqual(File.toPath("test.txt", requireExists=True), "test.txt")
        self.assertEqual(File.toPath("tEst.txt", requireExists=True), "tEst.txt")

    def test_exists(self):
        self.assertFalse(File.exists(File.getAbsolutePath("doesntExist")))

        self.assertFalse(File.exists(Path("doesntExist")))
        self.assertFalse(File.exists("doesntExist"))
        self.assertFalse(File.exists("doesntExist.txt"))

        File.write("test.txt")
        self.assertTrue(File.exists("test.txt"))
        self.assertTrue(File.exists("tEst.txt"))
        File.delete("test.txt")
        self.assertFalse(File.exists("test.txt"))
        self.assertFalse(File.exists("Test.txt"))

        File.createFolder("test")
        self.assertTrue(File.exists("test"))
        self.assertTrue(File.exists("teSt"))
        File.delete("test")
        self.assertFalse(File.exists("test"))

    def test_getWorkingDir(self):
        workingDir = File.getWorkingDir()
        File.setWorkingDir("folder")
        self.assertEqual(File.getWorkingDir().removeFromStart(workingDir), "folder")
        File.setWorkingDir(workingDir)

    def test_setWorkingDir(self):
        workingDir = File.getWorkingDir()
        self.assertRaises(Exception, File.setWorkingDir, "file.txt")
        self.assertEqual(File.setWorkingDir("folder"), File.getAbsolutePath(Path("")))
        File.setWorkingDir(workingDir)

    def test_getAbsolutePath(self):
        path = Path("folder")
        self.assertEqual(File.getAbsolutePath(path), File.getWorkingDir().addPath("folder"))
        path = Path("folder/file.txt")
        self.assertEqual(File.getAbsolutePath(path, "folder"), File.getWorkingDir().addPath("folder/folder/file.txt"))
        self.assertTrue(File.sameDestination(File.getAbsolutePath(path, "foldEr"), File.getWorkingDir().addPath("folder/folder/file.txt")))
        self.assertEqual(File.getAbsolutePath(File.getAbsolutePath(Path("test.txt")), "folder"), File.getAbsolutePath("test.txt"))

    def test_getRelativePath(self):
        self.assertRaises(Exception, File.getRelativePath, "test", "basePathWithFile.txt")
        self.assertRaises(AttributeError, File.getRelativePath, "folder/file.txt", "anotherFolder")

        path = File.getAbsolutePath(Path("folder/folder2/file.txt"))
        self.assertEqual(File.getRelativePath(path), "folder/folder2/file.txt")
        self.assertEqual(File.getRelativePath(path, "folder"), "folder2/file.txt")
        self.assertEqual(File.getRelativePath(path, "folder/folder2"), "file.txt")
        self.assertEqual(File.getRelativePath(path, "folder/foldEr2"), "file.txt")

    def test_sameDestination(self):
        self.assertTrue(File.sameDestination("folder", "folder"))
        self.assertTrue(File.sameDestination("Folder", "folder"))
        self.assertTrue(File.sameDestination("Folder", "foldeR"))
        self.assertTrue(File.sameDestination("file.txt", "file.txt"))
        self.assertTrue(File.sameDestination("File.txt", "file.txt"))
        self.assertTrue(File.sameDestination("file.txt", "File.txt"))
        self.assertTrue(File.sameDestination(File.getAbsolutePath("file.txt"), File.getAbsolutePath("file.txt")))
        self.assertTrue(File.sameDestination("file.txt", File.getAbsolutePath("file.txt")))
        self.assertTrue(File.sameDestination(File.getAbsolutePath("file.txt"), "file.txt"))
        self.assertTrue(File.sameDestination(File.getAbsolutePath("folder"), File.getAbsolutePath("folder")))
        self.assertTrue(File.sameDestination("folder", File.getAbsolutePath("folder")))
        self.assertTrue(File.sameDestination(File.getAbsolutePath("folder"), "folder"))

        self.assertFalse(File.sameDestination(File.getAbsolutePath("folder"), "folder.txt"))
        self.assertFalse(File.sameDestination("folder/file.txt", "folder"))
        self.assertFalse(File.sameDestination(File.getAbsolutePath("folder", "folder"), "folder"))
        self.assertFalse(File.sameDestination(File.getAbsolutePath("foldEr", "folder"), "fOlder"))

    def test__read_txt(self):
        File.write("test.txt")
        with open("test.txt", "r") as textIO:
            self.assertIsNone(File._read_txt(textIO))
        
        File.write("test2.txt", ["hello", "there", 2])
        with open("test2.txt", "r") as textIO:
            self.assertEqual(File._read_txt(textIO), ["hello", "there", 2])

    def test__write_txt(self):
        with open("test.txt", "w") as textIO:
            self.assertIsNone(File._write_txt(textIO, None))

        with open("test.txt", "w") as textIO:
            self.assertEqual(File._write_txt(textIO, ["hello", "there", 2]), '["hello", "there", 2]')

    def test_read(self):
        self.assertRaises(Exception, File.read, "folder")
        self.assertRaises(Exception, File.read, File.getAbsolutePath("folder"))
        self.assertRaises(EnvironmentError, File.read, "missingMethod.nonExistantFiletype")
        self.assertRaises(EnvironmentError, File.read, File.getAbsolutePath("missingMethod.nonExistantFiletype"))

        self.assertEqual(File.read(File.getAbsolutePath("doesntExist.txt")), None)
        self.assertEqual(File.read("doesntExist.txt"), None)
        self.assertEqual(File.read(File.getAbsolutePath("doesntExist.txt"), "default"), "default")
        self.assertEqual(File.read("doesntExist.txt", "default"), "default")

        File.write("test.txt", "hello")
        self.assertEqual(File.read("test.txt"), "hello")

        File.write("test2.txt", ["hello", "there", 2])
        self.assertEqual(File.read("test2.txt"), ["hello", "there", 2])

        File.write(File.getAbsolutePath("test3.txt"), ["hello", "there", 2])
        self.assertEqual(File.read("test3.txt"), ["hello", "there", 2])

        File.write(File.getAbsolutePath("test3.txt"), ["hello", "there", 2], overwrite=True)
        self.assertEqual(File.read("tEst3.txt"), ["hello", "there", 2])

    def test_write(self):
        self.assertRaises(Exception, File.write, "folder")
        self.assertRaises(Exception, File.write, File.getAbsolutePath("folder"))
        self.assertRaises(EnvironmentError, File.write, "missingMethod.nonExistantFiletype")
        self.assertIsNone(File.write("test.txt"))
        self.assertRaises(FileExistsError, File.write, "test.txt")
        self.assertRaises(FileExistsError, File.write, "tEst.txt")
        self.assertRaises(FileExistsError, File.write, File.getAbsolutePath("test.txt"))

        self.assertEqual(File.write("folder/folder/folder/test.txt", "hello"), '"hello"')
        self.assertEqual(File.write("folder/folder/folder/test.txt", "helloa", overwrite=True), '"helloa"')
        self.assertEqual(File.write(File.getAbsolutePath("folder/folder/folder/test.txt"), "hellos", overwrite=True), '"hellos"')

    def test_rename(self):
        File.write("folder/test.txt")
        self.assertRaises(NameError, File.rename, "folder/test.txt", "aux")
        self.assertRaises(NameError, File.rename, "folder", "aux")

        File.rename("folder/test.txt", "hello")
        self.assertTrue(File.exists("folder/hello.txt"))
        self.assertFalse(File.exists("folder/test.txt"))
        File.rename("folder", "folder2")
        self.assertTrue(File.exists("folder2"))
        self.assertFalse(File.exists("folder"))


    def test_copy(self):
        File.write("exists.txt", 1)
        File.write("exists2.txt", 2)
        File.write("folder/exists.txt", 3)
        self.assertRaises(Exception, File.copy, "doesntExist.txt", "")
        self.assertRaises(Exception, File.copy, "doesntExist", "")
        self.assertRaises(Exception, File.copy, File.getAbsolutePath("doesntExist"), "")
        self.assertRaises(RecursionError, File.copy, "exists.txt", "exists.txt")
        self.assertRaises(RecursionError, File.copy, File.getAbsolutePath("exists.txt"), "exists.txt")
        self.assertRaises(AttributeError, File.copy, "exists.txt", "file.tsv")
        self.assertRaises(FileExistsError, File.copy, "exists.txt", "exists2.txt")
        self.assertRaises(FileExistsError, File.copy, "exists.txt", File.getAbsolutePath("exists2.txt"))
        self.assertRaises(NotADirectoryError, File.copy, "folder", "exists.txt")
        self.assertRaises(FileExistsError, File.copy, "exists.txt", "folder")
        self.assertRaises(FileExistsError, File.copy, "folder", "")

        self.assertRaises(NameError, File.copy, "exists.txt", "aux.txt")
        self.assertRaises(NameError, File.copy, "exists.txt", "aux")
        self.assertRaises(NameError, File.copy, "folder", "aux")

        File.copy(File.getAbsolutePath("Exists.txt"), "exists7.txt")
        self.assertEqual(File.read("exisTs7.txt"), 1)

        File.copy(File.getAbsolutePath("exists.txt"), "exists6.txt")
        self.assertEqual(File.read("exists6.txt"), 1)

        File.copy("exists.txt", "exists3.txt")
        self.assertEqual(File.read("exists3.txt"), 1)

        File.copy("exists2.txt", "folder")
        self.assertEqual(File.read("folder/exists2.txt"), 2)

        File.copy("exists2.txt", File.getAbsolutePath("folder2"))
        self.assertEqual(File.read("folder2/exists2.txt"), 2)

        File.copy("exists2.txt", "newFolder")
        self.assertEqual(File.read("newFolder/exists2.txt"), 2)

        File.copy(File.getAbsolutePath("exists2.txt"), File.getAbsolutePath("newFolder2"))
        self.assertEqual(File.read("newFolder2/exists2.txt"), 2)

        File.copy("folder", "folder4")
        self.assertEqual(File.read("folder4/exists.txt"), 3)
        self.assertEqual(File.read("folder4/exists2.txt"), 2)

        File.write("folder3/exists4.txt", 4)
        File.write("folder3/exists5.txt", 5)
        File.copy("folder3", "")
        self.assertEqual(File.read("exists4.txt"), 4)

        File.copy("exists.txt", "exists2.txt", overwrite=True)
        self.assertEqual(File.read("exists2.txt"), 1)

        File.copy("exists.txt", "folder", overwrite=True)
        self.assertEqual(File.read("folder/exists.txt"), 1)

        File.copy("exists.txt", "folDer9")
        self.assertEqual(File.read("folder9/exists.txt"), 1)

        File.copy(File.getAbsolutePath("exists.txt"), File.getAbsolutePath("folder3"), overwrite=True)
        self.assertEqual(File.read("folder3/exists.txt"), 1)

        File.copy("folder3", "", overwrite=True)
        self.assertEqual(File.read("exists4.txt"), 4)
        self.assertEqual(File.read("exists5.txt"), 5)

    def test_createFolder(self):
        self.assertFalse(File.createFolder(""))
        self.assertTrue(File.createFolder("folder"))
        self.assertFalse(File.createFolder("folder"))
        self.assertFalse(File.createFolder("Folder"))
        self.assertTrue(File.createFolder("folder/folder/folder/file.txt"))
        self.assertTrue(File.createFolder(File.getAbsolutePath("folder/folder/folder2")))
        self.assertTrue(File.createFolder(File.getAbsolutePath("folder/folder/Folder3")))

    def test_clearFolder(self):
        File.write("folder/test.txt", 5)
        self.assertTrue(File.clearFolder("folder"))
        self.assertEqual(len(File.getPaths("folder")), 0)

        File.write("folder/test.txt", 5)
        self.assertTrue(File.clearFolder("Folder"))
        self.assertEqual(len(File.getPaths("folDer")), 0)

        File.write("folder/test.txt", 5)
        File.write("folder/folder2/tes2.txt", 3)
        self.assertTrue(File.clearFolder("folder/test.txt"))
        self.assertEqual(len(File.getPaths("folder")), 0)

        File.write("folder/test.txt", 5)
        File.write("folder/folder2/tes2.txt", 3)
        self.assertTrue(File.clearFolder(File.getAbsolutePath("folder/test.txt")))
        self.assertEqual(len(File.getPaths("folder")), 0)

    def test_trash(self):
        self.assertFalse(File.trash("doesntExist.txt"))
        self.assertFalse(File.trash("doesntExist"))
        self.assertFalse(File.trash(File.getAbsolutePath("doesntExist")))

        File.write("exists.txt")
        self.assertTrue(File.trash("exists.txt"))

        File.write("folder/exists.txt")
        self.assertTrue(File.trash("folder"))

        File.write("folder/exists.txt")
        self.assertTrue(File.trash(File.getAbsolutePath("folder")))

        File.write("folder/exists.txt")
        self.assertTrue(File.trash(File.getAbsolutePath("Folder")))
        self.assertEqual(len(File.getPaths("folder")), 0)

    def test_delete(self):
        self.assertFalse(File.delete("doesntExist.txt"))
        self.assertFalse(File.delete("doesntExist"))
        self.assertFalse(File.delete(File.getAbsolutePath("doesntExist")))

        File.write("exists.txt")
        self.assertTrue(File.delete("exists.txt"))

        File.write("folder/exists.txt")
        self.assertTrue(File.delete("folder"))

        File.write("folder/exists.txt")
        self.assertTrue(File.delete(File.getAbsolutePath("folder")))

        File.write("folder/exists.txt")
        self.assertTrue(File.delete(File.getAbsolutePath("Folder")))
        self.assertEqual(len(File.getPaths("folder")), 0)

    def test_getPaths(self):
        File.write("test.txt")
        File.write("folder/test2.txt")
        File.write("folder/test3.txt")

        self.assertEqual(len(File.getPaths("doesntExist")), 0)
        self.assertEqual(len(File.getPaths("doesntExist/doesntExist.txt")), 0)

        self.assertEqual(len(File.getPaths()), 4)
        self.assertEqual(len(File.getPaths("")), 4)
        self.assertEqual(len(File.getPaths("doesntExist.txt")), 4)
        self.assertEqual(len(File.getPaths(File.getAbsolutePath(""))), 4)

        self.assertEqual(len(File.getPaths("folder/test2.txt")), 2)
        self.assertEqual(len(File.getPaths(File.getAbsolutePath("folder/test2.txt"))), 2)
        self.assertEqual(len(File.getPaths("folder")), 2)
        self.assertEqual(len(File.getPaths(File.getAbsolutePath("folder"))), 2)
        self.assertEqual(len(File.getPaths(File.getAbsolutePath("folDer"))), 2)

    def test_getTimeModified(self):
        File.write("folder/test.txt")
        # Sometimes it got a really small negative number because it's so fast, such as "-2.384185791015625e-07"
        self.assertTrue(-0.1 <= time.time() - File.getTimeModified("folder/test.txt") < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeModified(File.getAbsolutePath("folder/test.txt")) < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeModified("folder") < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeModified(File.getAbsolutePath("folder")) < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeModified("Folder") < 0.1)

    def test_getTimeCreated(self):
        File.write("folder/test.txt")
        self.assertTrue(-0.1 <= time.time() - File.getTimeCreated("folder/test.txt") < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeCreated(File.getAbsolutePath("folder/test.txt")) < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeCreated("folder") < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeCreated(File.getAbsolutePath("folder")) < 0.1)
        self.assertTrue(-0.1 <= time.time() - File.getTimeCreated("Folder") < 0.1)

    def test_threads(self):
        threads = []
        queue = mp.Queue()
        for i in range(count := 10):
            threads.append(mp.Process(target=threadTest, args=(queue, i)))
        for thread in threads:
            thread.start()

        results = []
        for i in range(count):
            self.assertNotIn(get := queue.get(), results)
            results.append(get)

        self.assertEqual(len(results), count)

def threadTest(queue, i):
    """
    Simple target function for multiprocess testing

    :param mp.Queue queue:
    :param int i:
    """
    queue.put(int(File.write("test.txt", i, overwrite=True)))

if __name__ == "__main__":
    x = unittest.main()







    # @staticmethod
    # def readValue(path, key):
    #     """
    #
    #     :param path:
    #     :param key:
    #     :return:
    #     """
    #     path = File.toPath(path)
    #     values = File.read(path, default={})
    #     if key not in values:
    #         return None
    #     return values[key]
    #
    # @staticmethod
    # def setValue(path, key, value):
    #     path = File.toPath(path)
    #     """
    #     """
    #     values = File.read(path, default={})
    #     values[key] = value
    #     File.write(path, values)
    #     return value
    #
    # @staticmethod
    # def index(path, key=None, value=None):
    #     path = File.toPath(path)
    #     """
    #     """
    #     if value is not None:
    #         return File().setValue(path, key, value)
    #     else:
    #         if key is None:
    #             return File().read(path, default={})
    #         else:
    #             return File().readValue(path, key)




# class FileTSV(File):
#     def scrubDictOfDicts(self, dictOfDicts, indexName):
#         """
#         dictOfDicts can be singular dict also
#         """
#         if isinstance(dictOfDicts, dict):
#             if isinstance(lib.dictFirstValue(dictOfDicts), dict):
#                 return dictOfDicts
#             else:
#                 return {dictOfDicts[indexName]: dictOfDicts}
#
#         print(dictOfDicts)
#         lib.error("dictOfDicts failed scrubbing, printed above")
#
#     def tsvWrite(self, filepath, dictOfDicts, indexName):
#         dictOfDicts = self.scrubDictOfDicts(dictOfDicts, indexName)
#         filepath = self.yesTsv(filepath)
#         self.createFolderPath(filepath)
#
#         with open(filepath, 'w') as tsvfile:
#             writer = csv.writer(tsvfile, delimiter = "\t", lineterminator = "\n")
#             writer.writerow(list(lib.dictFirstValue(dictOfDicts, iterate = True).columns))
#             for index, subDict in dictOfDicts.items():
#                 writer.writerow(list(subDict.values()))
#         return dictOfDicts
#
#     def tsvAppend(self, filepath, dictOfDicts, indexName):
#         """
#         Write instead if file doesn't exist
#         """
#         filepath = self.yesTsv(filepath)
#
#         if not self.exists(filepath):
#             return self.tsvWrite(filepath, dictOfDicts, indexName)
#
#         dictOfDicts = self.scrubDictOfDicts(dictOfDicts, indexName)
#
#         with open(filepath, 'a') as tsvfile:
#             writer = csv.writer(tsvfile, delimiter = "\t", lineterminator = "\n")
#             for index, subDict in dictOfDicts.items():
#                 writer.writerow(list(subDict.values()))
#
#         return dictOfDicts
#
#     def tsvRowToDict(self, row):
#         return {k: lib.strToDynamicType(v) for k, v in row.items()}
#
#     def tsvRead(self, filepath, indexName):
#         """
#         Manually add index to each created dict
#         Check for indexName duplicates, if there are any just use the last one
#         row in reader is an iterator I think, it contains all values as strings, first row will be the labels
#         """
#         filepath = self.yesTsv(filepath)
#         if not self.exists(filepath):
#             return {}
#
#         with open(filepath, 'r') as tsvfile:
#             reader = csv.DictReader(tsvfile, delimiter = "\t")
#             returnDict = {subDict[indexName]: subDict for subDict in map(self.tsvRowToDict, reader)}
#             # returnDict = {subDict[indexName]: subDict for subDict in map(dict, reader)}  # Without casting, 3 times faster
#
#         return returnDict
#
#     def tsvUpdate(self, filepath, values, indexName):
#         """
#         See if row exists, insert new if it doesn't, update if it does
#         Use pandas dataframe?
#         """
#         df = pd.read_csv(filepath)
#         print(df.head(5))
#
#     def yesTsv(self, filepath):
#         # return "{}.tsv".format(self.noFiletypeEnding(filepath))
#         return ""



"""
Base component of generalfile
"""
from send2trash import send2trash
import os
import shutil
import json
import pathlib

from generallibrary.time import Timer, sleep
from generalfile.extensions.classfiletsv import FileTSV

class File(FileTSV):
    """
    File class exists so that FileTSV for example can inherit it.
    Method parameter 'path' takes a Path or a Str, Str is converted to Path.
    """
    timeoutSeconds = 5
    deadLockSeconds = 3
    def __init__(self):
        raise UserWarning("No need to instantiate File, all methods are static or classmethods")

    @staticmethod
    def toPath(path, requireFiletype=None, requireExists=None):
        """
        Makes sure we're using Path class.
        Has built-in parameters to scrub.
        Uses Path.toPath() but it also has extra functionality to check whether path exists.

        :param str path: Resemble a filepath or folderpath
        :param requireFiletype: True, False or filetype ("txt" or "tsv")
        :param bool requireExists: Require file or folder to exist or not
        :return: Same Path or a new Path if a string was given
        :raises AttributeError: if requireFiletype is True and it's not a file and vice versa
        :raises TypeError: if requireFiletype is a specified filetype and doesn't match filetype
        :raises FileNotFoundError: if requireExists is True and file doesn't exist
        :raises FileExistsError: if requireExists is False and file exists
        """
        path = Path.toPath(path, requireFiletype)

        if requireExists is not None:
            exists = File.exists(path)
            if exists and not requireExists:
                raise FileExistsError("Path exists", path)
            if requireExists and not exists:
                raise FileNotFoundError("Path doesn't exist", path)
        return path

    @staticmethod
    def exists(path):
        """
        Works for both file and folder.

        :param str path: Resemble a filepath or folderpath
        :return: Whether file or folder exists or not
        :rtype: bool
        """
        path = Path.toPath(path)
        path = File.getAbsolutePath(path)
        return os.path.exists(path)

    @staticmethod
    def getWorkingDir():
        """
        :returns: Current working directory as absolute Path
        """
        return Path(os.getcwd())

    @staticmethod
    def setWorkingDir(path):
        """
        Sets new working directory, if a relative path is given then current working directory will be taken into account.

        :param str path: Path to folder
        :return: Absolute Path to new working directory
        """
        path = File.toPath(path, requireFiletype=False)
        path = File.getAbsolutePath(path)
        File.createFolder(path)
        os.chdir(path)
        return path

    @staticmethod
    def getAbsolutePath(path, basePath=None):
        """
        Gets absolute path based on absolute basePath unless path already is absolute.

        :param str path: Generic Path to be converted.
        :param str basePath: Optional base Path to a folder. Is converted to absolute using current work dir if it's relative.
        :return: Absolute Path
        """
        path = File.toPath(path)
        if path.isAbsolute:
            return path

        if basePath is None:
            basePath = File.getWorkingDir()
        else:
            basePath = File.toPath(basePath, requireFiletype=False)
            basePath = File.getAbsolutePath(basePath)
        return basePath.addPath(path)

    @staticmethod
    def getRelativePath(path, basePath=None):
        """
        Gets relative path based on basePath.
        If basePath is None then current working dir will be used.

        :param str path: Generic Path to be converted
        :param str basePath: Optional base Path to a folder. Is converted to absolute using current work dir if it's relative.
        :return: Relative Path
        :raises AttributeError: If basePath isn't a part of given Path
        """
        path = File.toPath(path)
        path = File.getAbsolutePath(path)

        if basePath is None:
            basePath = File.getWorkingDir()
        else:
            basePath = File.toPath(basePath, requireFiletype=False)
            basePath = File.getAbsolutePath(basePath)
        if (relativePath := path.removeFromStart(basePath)).isAbsolute:
            raise AttributeError("Working directory ('{}') is not part of path ('{}') so we cannot get the relative path".format(File.getWorkingDir(), path))
        return relativePath

    @staticmethod
    def sameDestination(path1, path2):
        """
        Detects if two paths point to same folder or file based on current working directory by converting to absolute paths.
        Compares identifiers instead of strings directly, to ignore case.

        :param str path1: Generic Path
        :param str path2: Generic Path
        :return: Whether paths point to same destination or not.
        """
        path1 = File.toPath(path1)
        path2 = File.toPath(path2)

        path1 = File.getAbsolutePath(path1)
        path2 = File.getAbsolutePath(path2)

        return Path.identifier(path1) == Path.identifier(path2)

    @staticmethod
    def _read_txt(textIO):
        """
        Reads txt file from IO stream as JSON.
        This method is included in default File because they're needed for backup.

        :param io.TextIO textIO: From "with open(Path) as textIO"
        :return: serializable object or None
        """
        try:
            read = textIO.read()
        except:
            return None
        else:
            if read == "":
                return None
            return json.loads(read)

    @staticmethod
    def _write_txt(textIO, serializable):
        """
        Writes to txt file IO stream with JSON.
        This method is part of default File because they're needed for backup.

        :param io.TextIO textIO: From "with open(Path) as textIO"
        :param serializable: Any serializable object that json can use
        :return: json string (json.dumps)
        """
        jsonDumps = json.dumps(serializable)
        try:
            textIO.write(jsonDumps)
        except:
            return None
        else:
            if jsonDumps == "" or serializable is None:
                return None
            return jsonDumps

    @classmethod
    def read(cls, path, default=None, **kwargs):
        """
        Dynamic function that can read any document if methods exist for that filetype

        Generic read method:
            Only used dynamically by this method.

            Suffix has to match a filetype.

            Method has to take path as parameter and return a useful object for that filetype.

        txt:
            Takes: None

            Returns: Serializable object.

        tsv:
            Takes: (header: bool, column: bool) - Whether these index exists in file or not. Defaults to False.

            Returns: Pandas DataFrame

        :param str path: Path or Str
        :param default: What is returned as backup if reading fails or file doesn't exist
        :return: Useful object for that filetype (Dynamic)
        :raises EnvironmentError: if read method is missing for that filetype
        """
        path = File.toPath(path, requireFiletype=True)
        exists = File.exists(path)
        if (readMethod := getattr(cls, "_read_{}".format(path.filetype), None)) is None:
            raise EnvironmentError("Missing read method for filetype {}".format(path.filetype))
        if not exists:
            return default

        with open(path, "r") as textIO:
            read = readMethod(textIO, **kwargs)
        return default if read is None else read

    @classmethod
    def write(cls, path, writeObj=None, overwrite=False, **kwargs):
        """
        Dynamic function that can write to any document if methods exist for that filetype.

        Can only lock and backup if file exists.

        Generic write method:
            Only used dynamically by this method.

            Suffix has to match a filetype.

            Method has to take path and useful object as parameter, return object is optional

        txt:
            Takes: Serializable object.

            Returns: Text string that is written to txt file.

        tsv:
            Takes: Pandas DataFrame

            Returns: (header: bool, column: bool) - Whether these index exists in file or not.

        :param bool overwrite: Set to False if overwriting shouldn't be allowed
        :param str path: Path or Str
        :param writeObj: A useful object that the dynamic write method can use
        :return: Whatever the write method returns
        :raises EnvironmentError: if write method is missing for that filetype
        :raises FileExistsError: if overwriting when overwrite is set to False
        """
        path = File.toPath(path, requireFiletype=True)
        exists = File.exists(path)
        if (writeMethod := getattr(cls, "_write_{}".format(path.filetype), None)) is None:
            raise EnvironmentError("Missing write method for filetype {}".format(path.filetype))
        if exists and not overwrite:
            raise FileExistsError("Tried to overwrite {} when overwrite was False".format(path))

        if not exists:
            File.createFolder(path)

        pathNew = path.setSuffix("NEW")
        pathLock = path.setSuffix("LOCK")
        timer = Timer()

        while True:
            try:
                with open(pathLock, "x") as lockIO:
                    with open(pathNew, "w") as textIO:
                        writeReturn = writeMethod(textIO, writeObj, **kwargs)
                    File.delete(path)
                    File.rename(pathNew, path.filenamePure)
                    lockIO.close()
                    File.delete(pathLock)
            except FileExistsError as e:
                # PermissionError would have triggered if we couldn't delete lock
                File.delete(pathLock)
            except PermissionError as e:
                pass
            else:
                break
            if timer.seconds() > File.timeoutSeconds:
                raise TimeoutError(f"Couldn't open {pathLock} for writing")
        return writeReturn

    @staticmethod
    def rename(path, name):
        """
        Rename a file or folder. Cannot change filetype.

        :param str path: Generic path to folder or file that exists.
        :param name: New name of folder or file.
        :return: None
        :raises EnvironmentError: If new path exists
        :raises NameError: If used invalid name
        """
        path = File.toPath(path, requireExists=True)
        if path.isFile:
            newPath = path.setFilenamePure(name).setSuffix(None)
        else:
            newPath = path.getParent().addPath(name)
        if File.exists(newPath):
            raise EnvironmentError(f"New path {newPath} exists already")
        try:
            os.rename(path, newPath)
        except FileExistsError:
            raise NameError(f"{newPath} probably contains an invalid name such as CON, PRN, NUL or AUX")

    @staticmethod
    def copy(path, destPath, overwrite=False):
        """
        Copy and paste a file to file, file to folder or folder to folder.
        Creates new folders if needed.
        When copying folder it excludes the parent folder, simply add the folder name to the end of target Path if that's desired.

        :param str path: Path or Str
        :param str destPath: Path to folder or file
        :param bool overwrite: Allow overwriting or not
        :return: None
        :raises RecursionError: If paths are identical
        :raises FileExistsError: If trying to overwrite when not allowed
        :raises AttributeError: If filetypes don't match
        :raises NotADirectoryError: If trying to copy folder to file
        :raises NameError: If used invalid name
        """
        path = File.toPath(path, requireExists=True)
        path = File.getAbsolutePath(path)
        destPath = File.toPath(destPath)
        destPath = File.getAbsolutePath(destPath)

        if File.sameDestination(path, destPath):
            raise RecursionError("Identical paths")

        if path.isFile:
            if destPath.isFile:
                if path.filetype != destPath.filetype:
                    raise AttributeError("Filetypes don't match")
            elif destPath.isFolder:
                destPath = destPath.addPath(path.filenameFull)
            if File.exists(destPath):
                if not overwrite:
                    raise FileExistsError("Not allowed to overwrite")

            File.createFolder(destPath)
            try:
                shutil.copy(path, destPath, follow_symlinks=False)
            except FileNotFoundError:
                raise NameError(f"{destPath} probably contains an invalid name such as CON, PRN, NUL or AUX")

        elif path.isFolder and destPath.isFolder:
            if not overwrite:
                filePathList = File.getPaths(path).getFiles()
                relativePaths = filePathList.getRelative(path)
                absoluteDestPaths = relativePaths.getAbsolute(destPath)
                if any(absoluteDestPaths.exists()):
                    raise FileExistsError("Atleast one file exists and not allowed to overwrite")

            try:
                shutil.copytree(path, destPath, dirs_exist_ok=True)
            except NotADirectoryError:
                raise NameError(f"{destPath} probably contains an invalid name such as CON, PRN, NUL or AUX")
        else:
            raise NotADirectoryError("Cannot copy folder to file")

    @staticmethod
    def createFolder(path):
        """
        Create folder(s) for path.
        If a filepath is given then the filename is ignored.

        :param str path: Path or Str
        :return: Whether any folders were created or not
        :raises NameError: If used invalid name
        """
        path = File.toPath(path)
        path = File.getAbsolutePath(path)
        path = path.getPathWithoutFile()

        if File.exists(path):
            return False

        try:
            pathlib.Path(path).mkdir(parents=True, exist_ok=True)
        except NotADirectoryError:
            raise NameError(f"{path} probably contains an invalid name such as CON, PRN, NUL or AUX")

        return True

    @staticmethod
    def clearFolder(path, delete=False):
        """
        Make a target folder empty of everything.

        :param str path: Path or Str
        :param bool delete: Whether to delete files or put them in trash
        :return: Whether folder existed or not
        """
        path = File.toPath(path)
        path = File.getAbsolutePath(path)
        path = path.getPathWithoutFile()

        if not File.exists(path):
            return False
        if delete:
            File.delete(path)
        else:
            File.trash(path)
        File.createFolder(path)
        return True

    @staticmethod
    def trash(path):
        """
        Puts a file or folder in trashcan.

        :param str path: Path inside working directory
        :return: Whether path exists or not
        """
        path = File.toPath(path)
        path = File.getRelativePath(path)
        workingDir = File.getWorkingDir()

        if not File.exists(path):
            return False
        send2trash(path)
        # Reset working dir because send2trash can change it if it removed part of it
        if File.getWorkingDir() != workingDir:
            File.setWorkingDir(workingDir)
        return True

    @staticmethod
    def delete(path):
        """
        Deletes a file or folder, skipping trashcan

        :param str path: Path or Str
        :return: Whether path exists or not
        """
        path = File.toPath(path)
        if not File.exists(path):
            return False

        timer = Timer()
        if path.isFile:
            while True:
                try:
                    os.remove(path)
                except FileNotFoundError:
                    return False
                # Try again since it might just be being used by another process
                except PermissionError:
                    pass
                else:
                    break
                if timer.seconds() > File.timeoutSeconds:
                    raise TimeoutError(f"Couldn't delete {path}")
        elif path.isFolder:
            shutil.rmtree(path, ignore_errors=True)
        # If path is working dir then shutil.rmtree() only clears folder.
        if not File.sameDestination(path, File.getWorkingDir()):
            while File.exists(path):
                sleep(0.001)
        return True

    @staticmethod
    def getPaths(path = None, maxDepth=0):
        """
        Get a PathList obj from a path containing absolute Paths to both files and folders inside path.
        PathList extends list class with some convenient Path functionality.

        :param path: Path to folder. File is ignored. Default is Path("") which reads current work dir.
        :param maxDepth: Maximum folder depth, 0 means no limit.
        :return: PathList containing every absolute Paths in folder.
        """
        if path is None:
            path = Path("")
        path = File.toPath(path)
        path = path.getPathWithoutFile()
        if not File.exists(path):
            return PathList()

        path = File.getAbsolutePath(path)
        pathFoldersLen = len(path.foldersList)

        folderPathsToSearch = PathList(path)
        pathList = PathList()

        while folderPathsToSearch:
            folderPath = folderPathsToSearch[0]
            for subPath in os.listdir(folderPath):
                absoluteSubPath = folderPath.addPath(subPath)
                pathList.append(absoluteSubPath)
                if absoluteSubPath.isFile:
                    continue
                if maxDepth and len(absoluteSubPath.foldersList) - pathFoldersLen >= maxDepth:
                    continue
                folderPathsToSearch.append(absoluteSubPath)

            del folderPathsToSearch[0]
        return pathList

    @staticmethod
    def getTimeModified(path):
        """
        Gets time of when file (or folder?) was last modified

        :param str path: Path or Str
        :return: Time or None if file wasn't found
        """
        path = File.toPath(path)
        try:
            return os.path.getmtime(path)
        except FileNotFoundError:
            return None
        # except PermissionError:

    @staticmethod
    def getTimeCreated(path):
        """
        Gets datetime of when file (or folder?) was created.
        Can be innacurate it seems when re-creating files.

        :param str path: Path or Str
        :return: Datetime or None
        """
        path = File.toPath(path)
        try:
            return os.path.getctime(path)
        except FileNotFoundError:
            return None

    @staticmethod
    def openFolder(path):
        """
        Open file explorer on given path, files are ignored

        :param str path: Generic path that exists
        """
        path = File.toPath(path, requireExists=True).getPathWithoutFile()
        os.startfile(path)



from generalfile.base.classpath import Path
from generalfile.base.classpathlist import PathList




"""
Base component of generalfile.
Path is unaware and ignorant of environment.
"""

class Path(str):
    """
    Immutable.
    Path inherits string and makes manipulating paths easy.
    All path parameters used in File are converted to Paths.
    Three stages of scrubbing:
        1. Checking for illegal characters in Path:__new__
        2. Slashes fixed in Path:__new__
        3. Path:identifier returns a comparable path (lowering all characters for starters)
    """
    suffixDelimeter = "_"

    def __new__(cls, text=None):
        """
        Doesn't remove caps but methods should ignore it.

        Invalid words in windows, probably too expensive to check for them?
        ("CON", "PRN", "AUX", "NUL", "COM0", "LPT0", "COM1", "LPT1", "COM2", "LPT2", "COM3", "LPT3", "COM4"
        , "LPT4", "COM5", "LPT5", "COM6", "LPT6", "COM7", "LPT7", "COM8", "LPT8", "COM9", "LPT9")

        :param str text: Takes a string or Path resembling a filepath or folderpath
        :raises WindowsError: If invlaid character found
        """

        if text is None:
            text = ""

        # Simple invalid characters testing
        for character in tuple(text):
            if character in "<>\"|?*":
                raise WindowsError("Invalid character {} in {}".format(character, text))

        text = text.replace("\\", "/")
        if text.endswith("/"):
            text = text[0:-1]
        if text.startswith("/"):
            text = text[1:]

        return super().__new__(cls, text)

    def __init__(self, _=None):
        super().__init__()

        self.isAbsolute = ":" in self
        self.isRelative = not self.isAbsolute
        self.partsList = self.split("/")

        # Catch some simple mistakes

        if self.count(".") > 1:
            raise WindowsError("More than one dot in {}".format(self))
        for i, part in enumerate(self.partsList):
            if ":" in part:
                if i or (part.index(":") != 1 or part.count(":") > 1):
                    raise WindowsError(": in part #{} ({})".format(i, part), self)
            if i != len(self.partsList) - 1:
                if "." in part:
                    raise WindowsError(". in part that's not last #{} ({})".format(i, part), self)

        self.filenamePure = None
        self.filenameFull = None
        self.suffix = None
        self.filetype = None

        if "." in self.partsList[-1]:
            self._pathWithoutType = self.split(".")[0]

            splitDot = self.partsList[-1].split(".")

            self.filenameFull = self.partsList[-1]
            self.filetype = splitDot[1]
            self.foldersList = self.partsList[0:len(self.partsList) - 1]

            splitSuffix = splitDot[0].split(self.suffixDelimeter)
            self.filenamePure = splitSuffix[0]

            if self.suffixDelimeter in self.filenameFull:
                self.suffix = splitSuffix[1]

        else:
            self.foldersList = self.partsList

        self.isFile = self.filenamePure is not None
        self.isFolder = not self.isFile


    @staticmethod
    def identifier(pathPart):
        """
        :param str pathPart: A full path, partial part or even only a filetype.
        :return: A string that can be compared with other paths' identifiers to see if they point to the same file or folder.
        """
        return pathPart.lower()

    @staticmethod
    def toPath(path, requireFiletype = None):
        """
        Makes sure we're using Path class.
        Has built-in parameters to scrub.

        :param str path: Generic Path
        :param requireFiletype: True, False or filetype ("txt" or "tsv")
        :return: Same Path or a new Path if a string was given
        :raises ValueError: If neither a str or path is given
        :raises AttributeError: if requireFiletype is True and it's not a file and vice versa
        :raises TypeError: if requireFiletype is a specified filetype and doesn't match filetype
        """
        if isinstance(path, Path):  # Because Path[0:5] for example is still a Path, so instantiate again to scrub
            path = Path(path)
        elif isinstance(path, str):
            path = Path(path)
        else:
            raise ValueError("path has neither instance Path nor str, it's {}".format(type(path)))
        
        if requireFiletype is not None:
            if requireFiletype is True:
                if not path.isFile:
                    raise AttributeError("Not a file", path)
            elif requireFiletype is False:
                if path.isFile:
                    raise AttributeError("Is a file", path)
            elif Path.identifier(requireFiletype) != Path.identifier(path.filetype):
                raise TypeError("Mismatch filetype", requireFiletype, path.filetype)
        
        return path

    def addPath(self, path):
        """
        Simply add a path to this one.

        :param str path: Path to be added at the end.
        :return: New conjoined path.
        :raises FileExistsError: If first path is a file.
        :raises AttributeError: If second path is absolute.
        """
        path = self.toPath(path)

        if self.filetype:
            raise FileExistsError("Cannot add {} to {} because first is a file".format(path, self))
        if path.isAbsolute:
            raise AttributeError("Cannot add {} to {} because second is absolute".format(path, self))

        foldersList = self.foldersList.copy()
        foldersList.extend(path.foldersList)
        return path.setFoldersList(foldersList)

    def getPathWithoutFile(self):
        """
        Get a new Path without file.
        :return:
        """
        if not self.isFile:
            return self
        return self.removeFromEnd(self.filenameFull)

    def setPartsList(self, partsList):
        """
        Get a new Path with replaced parts.

        :param list[str] partsList: List containing path parts.
        :return: New path with new parts.
        """
        return Path("/".join(partsList))

    def setFoldersList(self, foldersList):
        """
        Get a new Path with replaced folder list.

        :param list[str] foldersList: List containing folder names.
        :return: New path with new folders.
        """
        if foldersList == self.foldersList:
            return self
        if not foldersList:
            foldersList = False
        return self._getPath(foldersList=foldersList)

    def setFilenamePure(self, filenamePure):
        """
        Get a new Path with replaced pure filename.

        :param str filenamePure: Filename without anything else.
        :return: New path with new pure filename.
        """
        if filenamePure == self.filenamePure:
            return self
        return self._getPath(filenamePure=filenamePure)

    def setSuffix(self, suffix):
        """
        Get a new Path with replaced suffix.

        :param any suffix: Suffix or None
        :return: New Path with or without suffix.
        """
        if suffix == self.suffix:
            return self
        if not suffix:
            suffix = False
        return self._getPath(suffix=suffix)

    def setFiletype(self, filetype):
        """
        Get a new Path with replaed filetype.

        :param filetype:
        :return:
        """
        if filetype == self.filetype:
            return self
        return self._getPath(filetype=filetype)

    def _getPath(self, foldersList = None, filenamePure = None, suffix = None, filetype = None):
        """
        Returns a new Path with optionally changed attributes, if none are changed then it returns self.

        :rtype: Path
        """
        if foldersList is None:
            foldersList = self.foldersList.copy()
        if filenamePure is None:
            filenamePure = self.filenamePure
        if suffix is None:
            suffix = self.suffix
        if filetype is None:
            filetype = self.filetype

        if foldersList == self.foldersList and filenamePure == self.filenamePure and suffix == self.suffix and filetype == self.filetype:
            return self

        if foldersList is False:
            foldersList = []

        if not filenamePure:
            if suffix:
                raise AttributeError("Cannot have suffix without filenamePure")
            if filetype:
                raise AttributeError("Cannot have filetype without filenamePure")
        else:
            if not filetype:
                raise AttributeError("Cannot have filenamePure without filetype")

        if filenamePure:
            suffixParts = [filenamePure]
            if suffix:
                suffixParts.append(suffix)
            filenamePureWithSuffix = self.suffixDelimeter.join(suffixParts)
            foldersList.append("{}.{}".format(filenamePureWithSuffix, filetype))
            return Path("/".join(foldersList))

        else:
            return Path("/".join(foldersList))

    def startsWithPath(self, path):
        """
        See if this Path starts with given path.
        Every part (Folder or File) is tested individually, so it's not tested like a typical string.

        :param str path: Generic path to check with.
        :returns: Whether it starts with that path or not.
        """
        path = Path.toPath(path)
        if len(path.partsList) > len(self.partsList):
            return False

        for i, pathPart in enumerate(path.partsList):
            selfPart = self.partsList[i]
            if Path.identifier(pathPart) != Path.identifier(selfPart):
                return False
        return True

    def endsWithPath(self, path):
        """
        See if this Path ends with given path.
        Every part (Folder or File) is tested individually, so it's not tested like a typical string.

        :param str path: Generic path to check with.
        :returns: Whether it ends with that path or not.
        """
        path = Path.toPath(path)
        if len(path.partsList) > len(self.partsList):
            return False

        for i, pathPart in enumerate(reversed(path.partsList)):
            selfPart = self.partsList[-i - 1]
            if Path.identifier(pathPart) != Path.identifier(selfPart):
                return False
        return True

    def removeFromStart(self, path):
        """
        Get a new Path with atleast one part removed from it's start.

        :param str path: Path that should be removed.
        :returns: New path with removed parts from start.
        """
        path = Path.toPath(path)
        if not self.startsWithPath(path):
            return self
        newPartsList = self.partsList[len(path.partsList):]
        return path.setPartsList(newPartsList)

    def removeFromEnd(self, path):
        """
        Get a new Path with atleast one part removed from it's end.

        :param str path: Path that should be removed.
        :returns: New path with removed parts from end.
        """
        path = Path.toPath(path)
        if not self.endsWithPath(path):
            return self
        newPartsList = self.partsList[:-len(path.partsList)]
        return path.setPartsList(newPartsList)

    def getAbsolute(self, basePath=None):
        """
        SHORTCUT for File.getAbsolutePath(path, basePath).

        Gets absolute path based on basePath unless path already is absolute.
        If basePath is None then current working dir will be used.

        :param str basePath: Optional base Path to a folder. Is converted to absolute using current work dir if it's relative.
        :return: Absolute Path
        """
        return File.getAbsolutePath(self, basePath=basePath)

    def getRelative(self, basePath=None):
        """
        SHORTCUT for File.getAbsolutePath(path, basePath).

        Gets relative path based on basePath.
        If basePath is None then current working dir will be used.

        :param str basePath: Optional base Path to a folder. Is converted to absolute using current work dir if it's relative.
        :return: Relative Path
        :raises AttributeError: If basePath isn't a part of given Path
        """
        return File.getRelativePath(self, basePath=basePath)

    def getParent(self, generations=1, basePath=None):
        """
        Get a Path to any parent folder of a Path.

        If a relative Path is given then a relative Path is returned if possible.
        Relative Paths are converted to Absolute if going past base folder.
        Filepaths' first parent is the folder it's in.

        :param int generations: Which parent to get, default is 1.
        :param str basePath: Used if given Path must be converted to Absolute.
        :return: Path to parent or None if no parent found
        """
        absolutePath = self.getAbsolute(basePath)
        for i in range(generations):
            if len(absolutePath.partsList) == 1:
                return None
            absolutePath = absolutePath.removeFromEnd(absolutePath.partsList[-1])

        if self.isAbsolute:
            return absolutePath
        try:
            return absolutePath.getRelative(basePath)
        except AttributeError:
            return absolutePath



from generalfile.base.classfile import File



"""
Tests for generalfile module

Exceptions raised directly by function should be specific Exceptions
Exceptions raised directly by indirect function should be generic Exception
"""
import unittest
from generalfile.base.classpath import Path
from generalfile.base.classfile import File
from generalfile.base.classpathlist import PathList

class PathListTest(unittest.TestCase):
    def test_pathList(self):
        self.assertRaises(TypeError, PathList, 1)
        self.assertRaises(TypeError, PathList, [1])
        self.assertRaises(TypeError, PathList, "file.txt", 1)
        self.assertRaises(TypeError, PathList, "file.txt", ["folder", True])
        self.assertRaises(TypeError, PathList, "file.txt", ["folder"], ["folder2", None])

        self.assertEqual(PathList("file.txt"), ["file.txt"])
        self.assertEqual(PathList("file.txt", "folder"), ["file.txt", "folder"])
        self.assertEqual(PathList("file.txt", ["folder"]), ["file.txt", "folder"])
        self.assertEqual(PathList("file.txt", ["folder", "folder2/file.txt"]), ["file.txt", "folder", "folder2/file.txt"])
        self.assertEqual(PathList([Path("test.txt").getAbsolute()], "file.txt", ["folder"]), [Path("test.txt").getAbsolute(), "file.txt", "folder"])

    def test_toPath(self):
        self.assertFalse(all([isinstance(path, Path) for path in PathList(["file.txt", "folder"])]))
        self.assertTrue(all([isinstance(path, Path) for path in PathList(["file.txt", "folder"]).toPath()]))
        self.assertTrue(all([isinstance(path, Path) for path in PathList([Path("file.txt"), "folder"]).toPath()]))
        self.assertTrue(all([isinstance(path, Path) for path in PathList([Path("file.txt"), Path("folder")]).toPath()]))

        pathList = PathList(["file.txt", "folder"])
        pathList.toPath()
        self.assertTrue(all([isinstance(path, Path) for path in pathList]))

    def test_getFolders(self):
        self.assertEqual(PathList(["file.txt", "file2.txt"]).getFolders(), [])
        self.assertEqual(PathList(["file.txt", "folder"]).getFolders(), ["folder"])
        self.assertEqual(PathList(["folder1", "folder2/folder3"]).getFolders(), ["folder1", "folder2/folder3"])

    def test_getFiles(self):
        self.assertEqual(PathList(["file.txt", "file2.txt"]).getFiles(), ["file.txt", "file2.txt"])
        self.assertEqual(PathList(["file.txt", "folder"]).getFiles(), ["file.txt"])
        self.assertEqual(PathList(["folder1", "folder2/folder3"]).getFiles(), [])

    def test_getRelative(self):
        self.assertEqual(PathList(Path("file.txt").getAbsolute()).getRelative(), ["file.txt"])
        self.assertEqual(PathList([Path("file.txt").getAbsolute(), "file2.txt"]).getRelative(), ["file.txt", "file2.txt"])

    def test_getAbsolute(self):
        self.assertEqual(PathList(Path("file.txt")).getAbsolute(), [Path("file.txt").getAbsolute()])
        self.assertEqual(PathList([Path("file.txt"), "file2.txt"]).getAbsolute(), [Path("file.txt").getAbsolute(), Path("file2.txt").getAbsolute()])

    def test_exists(self):
        File.write("exists.txt")
        self.assertEqual(PathList("doesntExist.txt", "exists.txt").exists(), (False, True))
        File.delete("exists.txt")



if __name__ == "__main__":
    unittest.main()


































"""
Tests for class FileXXX
"""
import unittest

from generalfile.base.classfile import File
from test.base.setUpWorkDir import SetUpWorkDir

class FileXXXTest(unittest.TestCase):
    def setUp(self):
        """Set working dir and clear folder"""
        SetUpWorkDir.activate()

    def test__write_XXX(self):
        pass

    def test__read_XXX(self):
        pass




"""
Tests for class FileTSV
"""
import unittest
import pandas as pd

from generalfile.base.classfile import File
from test.base.setUpWorkDir import SetUpWorkDir

class FileTSVTest(unittest.TestCase):
    def setUp(self):
        """Set working dir and clear folder"""
        SetUpWorkDir.activate()

    def compareFrames(self, df1, df2):
        self.assertEqual(df1.shape, df2.shape)
        self.assertTrue(df1.columns.equals(df2.columns))
        self.assertTrue(df1.index.equals(df2.index))

    def writeAndReadDF(self, df):
        header, column = File.write("df.tsv", df, overwrite=True)
        read = File.read("df.tsv", header=header, column=column)
        return read

    def appendAndReadDF(self, obj):
        File.write("df.tsv", pd.DataFrame(), overwrite=True)
        File.tsvAppend("df.tsv", obj)
        read = File.read("df.tsv", header=False, column=False)
        return read

    def doTestsOnDataFrame(self, obj):
        df = pd.DataFrame(obj)
        self.compareFrames(df, self.writeAndReadDF(df))
        self.compareFrames(df.T, self.writeAndReadDF(df.T))

    def test_tsvWriteAndRead(self):
        self.doTestsOnDataFrame({"a": {"color": "red", "value": 5}, "b": {"color": "blue", "value": 2}})
        self.doTestsOnDataFrame([{"color": "red", "value": 5}, {"color": "blue", "value": 2}])
        self.doTestsOnDataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        self.doTestsOnDataFrame([[1, 2, 3], [4, 5, 6]])

    def test_tsvAppend(self):
        df = pd.DataFrame([[1, 2, 3], [4, 5, 6]])
        self.compareFrames(df, self.appendAndReadDF([[1, 2, 3], [4, 5, 6]]))
        self.compareFrames(df, self.appendAndReadDF([{"a": 1, "b": 2, "c": 3}, {"d": 4, "e": 5, "f": 6}]))
        self.compareFrames(df, self.appendAndReadDF({1: {"b": 2, "c": 3}, 4: {"e": 5, "f": 6}}))
        self.compareFrames(df, self.appendAndReadDF({1: [2, 3], 4: [5, 6]}))

        df = pd.DataFrame(["hello"])
        self.compareFrames(df, self.appendAndReadDF("hello"))







































"""
Extension for File to handle tsv files
"""
import csv
import pandas as pd
from generallibrary.types import typeChecker
from generallibrary.iterables import getRows

class FileTSV:
    """
    Extension for File to handle tsv files
    """

    @staticmethod
    def _indexIsNamed(index):
        """
        Simple version to see if a DataFrame index is named or not
        :param index: DataFrame's index (columns or index)
        """
        if str(index[0]) == "0" or str(index[0]) == "1":
            return False
        else:
            return True

    @classmethod
    def _write_tsv(cls, textIO, df):
        """
        Can write a bunch of different DataFrames to a TSV file.
        Doesn't support advanced pandas functionality.
        Should work with: Keys, Index and Transposed (8 combinations)
        If DataFrame has both keys and index then cell A1 becomes NaN

        TODO: Make it order columns if there are any so that they line up with append.

        :param generalfile.base.classfile.File cls: File inherits FileTSV
        :param textIO: Tsv file
        :param pd.DataFrame df: Generic df
        """

        typeChecker(df, pd.DataFrame)
        path = cls.toPath(textIO.name)

        if df.empty:
            with open(path, "w"):
                pass
            return False, False

        useHeader = cls._indexIsNamed(df.columns)
        useIndex = cls._indexIsNamed(df.index)

        df.to_csv(path, sep="\t", header=useHeader, index=useIndex)

        return useHeader, useIndex

    @classmethod
    def _read_tsv_helper(cls, path, header, column):
        return pd.read_csv(path, sep="\t", header=header, index_col=column).convert_dtypes()

    @classmethod
    def _read_tsv(cls, textIO, header=False, column=False):
        """
        If any cell becomes NaN then the header and column parameters are overriden silently.

        Can read a bunch of different DataFrames to a TSV file.
        Doesn't support advanced pandas functionality.
        Should work with: Keys, Index, Transposed, Header, Column (32 combinations).
        DataFrame in file can have a NaN A1 cell.

        :param generalfile.base.classfile.File cls:
        :param textIO:
        :param bool header: Use headers or not, overriden if any top left cell is NaN
        :param bool column: Use columns or not, overriden if any top left cell is NaN
        :rtype: pd.DataFrame
        """

        path = cls.toPath(textIO.name)

        header = "infer" if header else None
        column = 0 if column else None

        try:
            df = cls._read_tsv_helper(path, header, column)
        except pd.errors.EmptyDataError:
            return pd.DataFrame()

        # Get rid of empty cell (Happens if file was written with header=True, column=True)
        headerFalseColumnFalse = pd.isna(df.iat[0, 0])
        headerFalseColumnTrue = pd.isna(df.index[0])
        headerTrueColumnFalse = str(df.columns[0]).startswith("Unnamed: ")
        if headerFalseColumnFalse or headerFalseColumnTrue or headerTrueColumnFalse:
            header = "infer"
            column = 0
            df = cls._read_tsv_helper(path, header, column)
        else:
            # Get rid of name in index (Happens if file doesn't have an index and column=True)
            # Doesn't happen other way around for some reason, guess it's the internal order in pandas
            if df.index.name is not None:
                if header is None and column == 0:
                    df.index.rename(None, inplace=True)
                else:
                    header = None
                    column = None
                    df = cls._read_tsv_helper(path, header, column)

        if not cls._indexIsNamed(df.columns):
            df.columns = pd.RangeIndex(len(df.columns))
        if not cls._indexIsNamed(df.index):
            df.index = pd.RangeIndex(len(df.index))

        return df.convert_dtypes()

    @staticmethod
    def _tsvAppend_getRow(iterableObj, key=None):
        """
        Takes an object and returns a list of rows to use for appending.

        :param iterableObj: Iterable
        :param key: If iterableObj had a key to assigned it it's given here
        :return: A
        """
        row = [key] if key else []
        if isinstance(iterableObj, (list, tuple)):
            row.extend(iterableObj)
        elif isinstance(iterableObj, dict):
            for _, value in sorted(iterableObj.items()):
                row.append(value)
        return row

    @classmethod
    def tsvAppend(cls, path, obj):
        """
        Append an obj to the end of a TSV file.
        If a dict is given and there are iterables as values then the keys of the dict are the first value in each row.
        Otherwise keys in dicts are ignored.

        Identical append objects
         | [[1, 2, 3], [4, 5, 6]]
         | [{"a": 1, "b": 2, "c": 3}, {"d": 4, "e": 5, "f": 6}]
         | {1: {"b": 2, "c": 3}, 4: {"e": 5, "f": 6}}
         | {1: [2, 3], 4: [5, 6]}

        :param generalfile.base.classfile.File cls:
        :param path:
        :param obj: Iterable (Optionally inside another iterable) or a value for a single cell
        :raises AttributeError: If obj is empty
        """
        path = cls.toPath(path, requireFiletype="tsv", requireExists=True)
        if not obj and obj != 0:
            raise AttributeError("obj is empty")

        with open(path, 'a') as tsvfile:
            writer = csv.writer(tsvfile, delimiter = "\t", lineterminator = "\n")
            for row in getRows(obj):
                writer.writerow(row)
















































"""
Tests for generalfile module

Exceptions raised directly by function should be specific Exceptions
Exceptions raised directly by indirect function should be generic Exception
"""
import unittest

from generalfile.base.classpath import Path
from generalfile.base.classfile import File

class PathTest(unittest.TestCase):
    def test_path(self):
        self.assertRaises(WindowsError, Path, "tes*t.txt")
        self.assertRaises(WindowsError, Path, "te<st")
        self.assertRaises(WindowsError, Path, "te>st")
        self.assertRaises(WindowsError, Path, "te\"st")
        self.assertRaises(WindowsError, Path, "te|st")
        self.assertRaises(WindowsError, Path, "te?st")
        self.assertRaises(WindowsError, Path, "te:st")
        self.assertRaises(WindowsError, Path, "test.txt/test.txt")
        self.assertRaises(WindowsError, Path, "C:/folder/te:st.txt")
        self.assertRaises(WindowsError, Path, "C:/folder/te.st.txt")
        self.assertRaises(WindowsError, Path, "AB:/folder/test.txt")
        self.assertRaises(WindowsError, Path, ":A/folder/test.txt")

        self.assertEqual(Path("folder\\file.txt"), "folder/file.txt")
        self.assertEqual(Path("Folder/file.txt"), "Folder/file.txt")
        self.assertEqual(Path(""), "")
        self.assertEqual(Path("folder"), "folder")
        self.assertEqual(Path("folder/"), "folder")
        self.assertEqual(Path("folder\\"), "folder")
        self.assertEqual(Path("/folder"), "folder")
        self.assertEqual(Path("\\folder"), "folder")
        self.assertEqual(Path("file.txt"), "file.txt")

    def test_toPath(self):
        self.assertRaises(ValueError, Path.toPath, True)
        self.assertRaises(ValueError, Path.toPath, None)
        self.assertRaises(ValueError, Path.toPath, ["list"])

        self.assertRaises(AttributeError, Path.toPath, "file.txt", requireFiletype=False)
        self.assertRaises(AttributeError, Path.toPath, "folder", requireFiletype=True)

        self.assertRaises(TypeError, Path.toPath, "file.tsv", requireFiletype="txt")
        self.assertRaises(TypeError, Path.toPath, "folder/file.txt", requireFiletype="tsv")

        self.assertTrue(isinstance(Path.toPath("test"), Path))
        self.assertTrue(isinstance(Path.toPath("test/file.txt"), Path))
        self.assertTrue(isinstance(Path.toPath(""), Path))

    def test_addPath(self):
        self.assertRaises(FileExistsError, Path("file.txt").addPath, Path("folder"))
        self.assertRaises(FileExistsError, Path("file.txt").addPath, Path("file.txt"))

        self.assertRaises(AttributeError, Path("folder").addPath, Path("file.txt").getAbsolute())
        self.assertRaises(AttributeError, Path("folder").addPath, Path("file.txt").getAbsolute())

        self.assertEqual(Path("folder").addPath("file.txt"), "folder/file.txt")
        self.assertEqual(Path("folder").getAbsolute().addPath("file.txt").getRelative(), "folder/file.txt")
        self.assertEqual(Path("folder").addPath("folder2"), "folder/folder2")
        self.assertEqual(Path("folder/folder2").addPath("folder3/file.txt"), "folder/folder2/folder3/file.txt")

    def test_getPathWithoutFile(self):
        self.assertEqual(Path("test.txt").getPathWithoutFile(), "")
        self.assertEqual(Path("folder/test.txt").getPathWithoutFile(), "folder")
        self.assertEqual(Path("C:/folder/test.txt").getPathWithoutFile(), "C:/folder")
        self.assertEqual(Path("C:/folder/test_SUFFIX.tsv").getPathWithoutFile(), "C:/folder")

    def test_setPartsList(self):
        self.assertRaises(Exception, Path().setPartsList, ["test", "test.fail.txt"])
        self.assertRaises(Exception, Path().setPartsList, ["C:", "test.fail.txt"])
        self.assertRaises(Exception, Path().setPartsList, ["C:", "te:st.txt"])
        self.assertRaises(Exception, Path().setPartsList, ["file.txt", "test.txt"])

        self.assertEqual(Path("C:/test/file.txt").setPartsList(["folder"]), "folder")
        self.assertEqual(Path("").setPartsList(["folder", "test.txt"]), "folder/test.txt")
        self.assertEqual(Path("").setPartsList(["test.txt"]), "test.txt")

    def test_setFoldersList(self):
        self.assertRaises(Exception, Path().setFoldersList, ["test", "test.fail.txt"])
        self.assertRaises(Exception, Path().setFoldersList, ["C:", "test.fail.txt"])
        self.assertRaises(Exception, Path().setFoldersList, ["C:", "te:st.txt"])
        self.assertRaises(Exception, Path().setFoldersList, ["file.txt", "test.txt"])
        self.assertRaises(Exception, Path("folder/test.txt").setFoldersList, ["test.txt"])

        self.assertEqual(Path("").setFoldersList(["folder"]), "folder")
        self.assertEqual(Path("").setFoldersList(["folder", "test.txt"]), "folder/test.txt")
        self.assertEqual(Path("test.txt").setFoldersList(["folder"]), "folder/test.txt")
        self.assertEqual(Path("").setFoldersList(["test.txt"]), "test.txt")
        self.assertEqual(Path("folder/file.txt").setFoldersList(["C:"]), "C:/file.txt")

    def test_setFilenamePure(self):
        self.assertRaises(Exception, Path().setFilenamePure, "test.txt")
        self.assertRaises(Exception, Path("hello.txt").setFilenamePure, "test.txt")

        self.assertEqual(Path("hello.txt").setFilenamePure("file"), "file.txt")
        self.assertEqual(Path("folder/hello.txt").setFilenamePure("file"), "folder/file.txt")
        self.assertEqual(Path("folder/hello_SUFFIX.txt").setFilenamePure("file"), "folder/file_SUFFIX.txt")

    def test_setSuffix(self):
        self.assertRaises(Exception, Path().setSuffix, "suffix")
        self.assertRaises(Exception, Path("hello").setSuffix, "suffix")
        self.assertRaises(Exception, Path("hello.txt").setSuffix, "suffix.txt")

        self.assertEqual(Path("file.txt").setSuffix("suffix"), "file_suffix.txt")
        self.assertEqual(Path("folder/file.txt").setSuffix("suffix"), "folder/file_suffix.txt")
        self.assertEqual(Path("folder/hello_SUFFIX.txt").setSuffix("test"), "folder/hello_test.txt")
        self.assertEqual(Path("C:/folder/hello_SUFFIX.txt").setSuffix("test"), "C:/folder/hello_test.txt")

    def test_setFiletype(self):
        self.assertRaises(Exception, Path("folder").setFiletype, "txt")
        self.assertRaises(Exception, Path().setFiletype, "txt")
        self.assertRaises(Exception, Path().getAbsolute().setFiletype, "txt")
        self.assertRaises(Exception, Path("test.txt").setFiletype, "")

        self.assertEqual(Path("test.tsv").setFiletype("txt"), "test.txt")
        self.assertEqual(Path("folder/test.tsv").setFiletype("txt"), "folder/test.txt")
        self.assertEqual(Path("folder/test_SUFFIX.tsv").setFiletype("txt"), "folder/test_SUFFIX.txt")
        self.assertEqual(Path("folder/test_SUFFIX.tsv").getAbsolute().setFiletype("txt").getRelative(), "folder/test_SUFFIX.txt")

    def test_startsWithPath(self):
        self.assertFalse(Path("file.txt").startsWithPath("folder"))
        self.assertFalse(Path("file.txt").startsWithPath("file"))
        self.assertFalse(Path("folder/file.txt").startsWithPath("file.txt"))
        self.assertFalse(Path("folder/file.txt").getAbsolute().startsWithPath("folder"))
        self.assertFalse(Path("file.txt").startsWithPath(File.getWorkingDir()))

        self.assertTrue(Path("folder/file.txt").startsWithPath("folder"))
        self.assertTrue(Path("file.txt").startsWithPath("file.txt"))
        self.assertTrue(Path("file_SUFFIX.txt").startsWithPath("file_SUFFIX.txt"))
        self.assertTrue(Path("filE.txt").startsWithPath("file.txt"))
        self.assertTrue(Path("file.txt").getAbsolute().startsWithPath(File.getWorkingDir()))

    def test_endsWithPath(self):
        self.assertFalse(Path("file.txt").endsWithPath("folder"))
        self.assertFalse(Path("file.txt").endsWithPath("file"))
        self.assertFalse(Path("folder/file.txt").endsWithPath("folder"))
        self.assertFalse(Path("folder/file.txt").getAbsolute().endsWithPath("file"))

        self.assertTrue(Path("folder/file.txt").endsWithPath("file.txt"))
        self.assertTrue(Path("file.txt").endsWithPath("file.txt"))
        self.assertTrue(Path("file_SUFFIX.txt").endsWithPath("file_SUFFIX.txt"))
        self.assertTrue(Path("filE.txt").endsWithPath("file.txt"))
        self.assertTrue(Path("filE.txt").getAbsolute().endsWithPath("file.txt"))

    def test_removeFromStart(self):
        self.assertEqual(Path("test.txt").removeFromStart("test.txt"), "")
        self.assertEqual(Path("folder/test.txt").removeFromStart("Folder"), "test.txt")
        self.assertEqual(Path("folder/test.txt").removeFromStart("test"), "folder/test.txt")
        self.assertEqual(Path("folder/test.txt").getAbsolute().removeFromStart(File.getWorkingDir()), "folder/test.txt")

    def test_removeFromEnd(self):
        self.assertEqual(Path("test.txt").removeFromEnd("test.txt"), "")
        self.assertEqual(Path("test.txt").removeFromEnd("txt"), "test.txt")
        self.assertEqual(Path("folder/test.txt").removeFromEnd("test.txt"), "folder")
        self.assertEqual(Path("folder/test.txt").removeFromEnd("test"), "folder/test.txt")
        self.assertEqual(Path("folder/test.txt").getAbsolute().removeFromEnd("folder/test.txt"), File.getWorkingDir())

    def test_getAbsolute(self):
        self.assertEqual(Path("").getAbsolute(), File.getWorkingDir())
        self.assertEqual(Path("test.txt").getAbsolute(), File.getWorkingDir().addPath("test.txt"))
        self.assertEqual(Path("test.txt").getAbsolute("folder"), Path("folder/test.txt").getAbsolute())

    def test_getRelative(self):
        self.assertRaises(Exception, Path("file.txt").getRelative, "doesntExist")
        self.assertEqual(Path("folder/test.txt").getRelative("folder"), "test.txt")
        self.assertEqual(Path("folder/test.txt").getAbsolute().getRelative("folder"), "test.txt")
        self.assertEqual(Path().getAbsolute().getRelative(), "")

    def test_getParent(self):
        self.assertEqual(Path("folder/file.txt").getParent(), "folder")
        self.assertEqual(Path("folder/file.txt").getParent(2), "")
        self.assertEqual(Path("folder/file.txt").getParent(2), File.getWorkingDir().getRelative())
        self.assertEqual(Path("folder/file.txt").getParent(0), "folder/file.txt")
        self.assertEqual(Path("folder/file.txt").getParent(3), File.getWorkingDir().getParent(1))

if __name__ == "__main__":
    unittest.main()



































from setuptools import setup, find_packages

from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="generalfile",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rickard "Mandera" Abraham',
    url="https://github.com/Mandera/generalfile",
    version="1.0.3",
    description=(
        "Easily manage windows files."
    ),
    packages=find_packages(),
    install_requires=["wheel"],
    classifiers=[
        "Operating System :: Microsoft :: Windows :: Windows 7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Desktop Environment :: File Managers",
    ]
)



"""
Base component of generalfile
"""

class PathList(list):
    """
    List that contains path with convenient methods.
    Takes None, paths or lists of paths as parameter for initializing.
    """
    def __init__(self, *args):
        paths = []
        for arg in args:
            if isinstance(arg, str):
                paths.append(arg)
            elif isinstance(arg, list) or isinstance(arg, tuple):
                for subArg in arg:
                    if not isinstance(subArg, str):
                        raise TypeError("Iterable doesn't contain a string")
                    paths.append(subArg)
            else:
                raise TypeError("Given arg is neither a string nor list or tuple")

        super().__init__(paths)

    def toPath(self):
        """
        Makes sure all paths inside are using class Path.
        Call this method when a Path's attribute or method is required.

        :return: Itself but all items are Paths class.
        """
        for path in self:
            if not isinstance(path, Path):
                break
        else:
            return self
        pathList = PathList([Path.toPath(path) for path in self])
        self.clear()
        for path in pathList:
            self.append(path)
        return self

    def getFolders(self):
        """
        :return: A new PathList only containing the folder paths.
        """
        return PathList([path for path in self.toPath() if path.isFolder])

    def getFiles(self):
        """
        :return: A new PathList only containing the file paths.
        """
        return PathList([path for path in self.toPath() if path.isFile])

    def getRelative(self, basePath=None):
        """
        :param basePath: Optional base path. Defaults to current work dir.
        :return: A new PathList with all paths converted to relative.
        """
        return PathList([File.getRelativePath(path, basePath) for path in self])

    def getAbsolute(self, basePath=None):
        """
        :param basePath: Optional base path. Defaults to current work dir.
        :return: A new PathList with all paths converted to absolute.
        """
        return PathList([File.getAbsolutePath(path, basePath) for path in self])

    def exists(self):
        """
        See which paths exists. Use all() or any() on result.
        :return: A tuple with Booleans
        :rtype: tuple[bool]
        """
        return tuple(File.exists(path) for path in self)




from generalfile.base.classpath import Path
from generalfile.base.classfile import File

